package org.qatools.factory;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.qatools.util.IntUtil;
import org.qatools.util.ListUtil;

public class AllPairsHelper extends QAToolsFactory {
    private static Logger logger = Logger.getLogger(AllPairsHelper.class);

  // --------------------------------------------------------
  // ----- allPairsIndexes
  // --------------------------------------------------------

    /**
     * 
     * @param list
     * @param list2
     * @return
     */
    protected static <A extends Object, B extends Object, C extends Object, D extends Object> int[][] allPairsIndexes(List<A> list, List<B> list2, List<C> list3, List<D> list4) {
    	if (list==null || list.size()==0 || list2==null || list2.size()==0 || list3==null || list3.size()==0) {
    		throw new RuntimeException("Need at least 3 parameters to create allPairs");
    	}
    	
    	//
    	// all the pairings
    	Map<Integer, List<Integer>> ab = IndexesToMap(list, list2);
    	Map<Integer, List<Integer>> ac = IndexesToMap(list, list3);
    	Map<Integer, List<Integer>> ad = IndexesToMap(list, list4);

    	Map<Integer, List<Integer>> ba = IndexesToMap(list2, list);
    	Map<Integer, List<Integer>> bc = IndexesToMap(list2, list3);
    	Map<Integer, List<Integer>> bd = IndexesToMap(list2, list4);

    	Map<Integer, List<Integer>> ca = IndexesToMap(list3, list);
    	Map<Integer, List<Integer>> cb = IndexesToMap(list3, list2);
    	Map<Integer, List<Integer>> cd = IndexesToMap(list3, list4);

    	Map<Integer, List<Integer>> da = IndexesToMap(list4, list);
    	Map<Integer, List<Integer>> db = IndexesToMap(list4, list2);
    	Map<Integer, List<Integer>> dc = IndexesToMap(list4, list3);
    	
    	//
    	// init array
    	int cols = (list4==null || 0==list4.size() ? 3 : 4);
    	int maxRows = list.size() * list2.size() * list3.size();    	
    	maxRows = (cols==3 ? maxRows : maxRows*list4.size());
    	int[][] indexes = new int[maxRows][cols];
    	logger.info("cols = " + cols);
    	logger.info("maxRows = " + maxRows);

    	//
    	// fill up the array
    	int actualRows=0;
    	while (0 < numActiveMaps(ab,ac,ad, ba,bc,bd, ca,cb,cd, da,db,dc) && actualRows < maxRows) {
    		//
    		// init new row
    		int a = -1; 
    		int b = -1; 
    		int c = -1; 
    		int d = -1;
    		int[] vals;
    		
    		//
    		// recalc new values
		    vals = doIt(a,b,c,d,  "a","b","c","d",  ab,ac,ad,  ba,ca,da, list);
		    a=vals[0];
		    b=vals[1];
		    c=vals[2];
		    d=vals[3];
		    
		    vals = doIt(b,a,c,d,  "b","a","c","d",  ba,bc,bd,  ab,cb,db, list2);
		    a=vals[1];
		    b=vals[0];
		    c=vals[2];
		    d=vals[3];

		    vals = doIt(c,a,b,d,  "c","a","b","d",  ca,cb,cd,  ac,bc,dc, list3);
		    a=vals[1];
		    b=vals[2];
		    c=vals[0];
		    d=vals[3];

		    vals = doIt(d,a,b,c,  "d","a","b","c",  da,db,dc,  ad,bd,cd, list4);
		    a=vals[1];
		    b=vals[2];
		    c=vals[3];
		    d=vals[0];
		    
		    //
		    // clean maps - little extra tidying since doIt misses cleanup when there's a newB, newC, newD 
		    remove(ab, a,b, "a","b");
		    remove(ac, a,c, "a","c");
		    remove(ad, a,d, "a","d");
		    remove(ba, b,a, "b","a");
		    remove(bc, b,c, "b","c");
		    remove(bd, b,d, "b","d");
		    remove(ca, c,a, "c","a");
		    remove(cb, c,b, "c","b");
		    remove(cd, c,d, "c","d");
		    remove(da, d,a, "d","a");
		    remove(db, d,b, "d","b");
		    remove(dc, d,c, "d","c");
		    
		    //
		    // new row
		    int[] bRow = new int[cols];
		    String abcd = "";
		    if (1 <= cols) { bRow[0] = a; abcd += a + ","; }
		    if (2 <= cols) { bRow[1] = b; abcd += b + ","; }
		    if (3 <= cols) { bRow[2] = c; abcd += c + ","; }
		    if (4 <= cols) { bRow[3] = d; abcd += d + ","; }
		    
		    logger.info("New row: " + actualRows + " = " + abcd);
		    indexes[actualRows] = bRow;
		    
		    //
		    // verify the indexes by checking againsts the lists - catches ArrayOutOfBoundExceptoins
		    if (list != null) { list.get(a); }
		    if (list2 != null) { list2.get(b); }
		    if (list3 != null) { list3.get(c); }
		    if (list4 != null) { list4.get(d); }
		    
		    
		    actualRows++;		    
		    logger.info("actualRows = " + actualRows);
    		//logger.info();
    	}	// while numActiveMaps
    	
    	return removeEmptyRows(actualRows, indexes);
    } // allPairsIndexes


    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------

    /**
     * If both lists are NOT null, get the pairing(indexes) and convert to Map
     * 
     * @param list
     * @param list2
     * @return
     */
    private static <A extends Object, B extends Object> Map<Integer, List<Integer>> IndexesToMap(List<A> list, List<B> list2) {
    	Map<Integer, List<Integer>> map = null;
    	try {
    		map = new HashMap<Integer, List<Integer>>();
    		
    		for (int a=0; a<list.size(); a++) {
    			List<Integer> l = new LinkedList<Integer>();
    			for (int b=0; b<list2.size(); b++) {
    				l.add(b);    				
    			}
    			map.put(a, l);
    		}
    		
    	} catch (Exception ex) {
    		return null;
    	}
    	return map;
    }
    
    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------

    private static int numActiveMaps(
    		Map<Integer, List<Integer>> ab,
    		Map<Integer, List<Integer>> ac,
    		Map<Integer, List<Integer>> ad,
    		Map<Integer, List<Integer>> ba,
    		Map<Integer, List<Integer>> bc,
    		Map<Integer, List<Integer>> bd,
    		Map<Integer, List<Integer>> ca,
    		Map<Integer, List<Integer>> cb,
    		Map<Integer, List<Integer>> cd,
    		Map<Integer, List<Integer>> da,
    		Map<Integer, List<Integer>> db,
    		Map<Integer, List<Integer>> dc
    		) {
    	int num = 0;
    	String active = "";
    	String inactive = "";
    	
    	num = (ab==null || ab.size()==0 ? num : num+1);
    	num = (ac==null || ac.size()==0 ? num : num+1);
    	num = (ad==null || ad.size()==0 ? num : num+1);
    	num = (ba==null || ba.size()==0 ? num : num+1);
    	num = (bc==null || bc.size()==0 ? num : num+1);
    	num = (bd==null || bd.size()==0 ? num : num+1);
    	num = (ca==null || ca.size()==0 ? num : num+1);
    	num = (cb==null || cb.size()==0 ? num : num+1);
    	num = (cd==null || cd.size()==0 ? num : num+1);
    	num = (da==null || da.size()==0 ? num : num+1);
    	num = (db==null || db.size()==0 ? num : num+1);
    	num = (dc==null || dc.size()==0 ? num : num+1);

    	active = (ab==null || ab.size()==0 ? active : active+"ab,");
    	active = (ac==null || ac.size()==0 ? active : active+"ac,");
    	active = (ad==null || ad.size()==0 ? active : active+"ad,");
    	active = (ba==null || ba.size()==0 ? active : active+"ba,");
    	active = (bc==null || bc.size()==0 ? active : active+"bc,");
    	active = (bd==null || bd.size()==0 ? active : active+"bd,");
    	active = (ca==null || ca.size()==0 ? active : active+"ca,");
    	active = (cb==null || cb.size()==0 ? active : active+"cb,");
    	active = (cd==null || cd.size()==0 ? active : active+"cd,");
    	active = (da==null || da.size()==0 ? active : active+"da,");
    	active = (db==null || db.size()==0 ? active : active+"db,");
    	active = (dc==null || dc.size()==0 ? active : active+"dc,");

    	inactive = (ab==null || ab.size()==0 ? inactive+"ab," : inactive);
    	inactive = (ac==null || ac.size()==0 ? inactive+"ac," : inactive);
    	inactive = (ad==null || ad.size()==0 ? inactive+"ad," : inactive);
    	inactive = (ba==null || ba.size()==0 ? inactive+"ba," : inactive);
    	inactive = (bc==null || bc.size()==0 ? inactive+"bc," : inactive);
    	inactive = (bd==null || bd.size()==0 ? inactive+"bd," : inactive);
    	inactive = (ca==null || ca.size()==0 ? inactive+"ca," : inactive);
    	inactive = (cb==null || cb.size()==0 ? inactive+"cb," : inactive);
    	inactive = (cd==null || cd.size()==0 ? inactive+"cd," : inactive);
    	inactive = (da==null || da.size()==0 ? inactive+"da," : inactive);
    	inactive = (db==null || db.size()==0 ? inactive+"db," : inactive);
    	inactive = (dc==null || dc.size()==0 ? inactive+"dc," : inactive);

    	logger.info("numActiveMaps = " + num);
    	logger.info("active = " + active);
    	logger.info("inactive = " + inactive);

    	printMap("ab", ab);
    	printMap("ac", ac);
    	printMap("ad", ad);
    	printMap("ba", ba);
    	printMap("bc", bc);
    	printMap("bd", bd);
    	printMap("ca", ca);
    	printMap("cb", cb);
    	printMap("cd", cd);
    	printMap("da", da);
    	printMap("db", db);
    	printMap("dc", dc);
    	
    	return num;
    }
    
    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------

    /**
     * 
     * @param a
     * @param b
     * @param c
     * @param d
     * @param ab
     * @param ac
     * @param ad
     * @param ba
     * @param ca
     * @param da
     * @return
     */
    private static <A extends Object> int[] doIt(
    		int a, 
    		int b, 
    		int c, 
    		int d,  
    		String aa,
    		String bb,
    		String cc,
    		String dd,
    		Map<Integer, List<Integer>> ab,
    		Map<Integer, List<Integer>> ac,
    		Map<Integer, List<Integer>> ad,  
    		Map<Integer, List<Integer>> ba,
    		Map<Integer, List<Integer>> ca,
    		Map<Integer, List<Integer>> da,
    		List<A> list) {
	    if (list==null || list.size()==0) {
	    	return new int[] { a, b, c, d };
	    }

	    if (a == -1) {    
	        if (b != -1 && ba!=null && ba.get(b) != null) {
	            a = getRandomValue(b, ba);
	            logger.info("Matching (" + aa + ":" + a + ") with (" + bb + ":" + b + ")");
	            //remove(ab, a,b, aa,bb);
	            //remove(ba, b,a, bb,aa);
	        } else if (c != -1 && ca!=null && ca.get(c) != null) {
	            a = getRandomValue(c, ca);
	            logger.info("Matching (" + aa + ":" + a + ") with (" + cc + ":" + c + ")");
	            //remove(ac, a,c, aa,cc);
	            //remove(ca, c,a, cc,aa);
	        } else if (d != -1 && da!=null && da.get(d) != null) {
	            a = getRandomValue(d, da);
	            logger.info("Matching (" + aa + ":" + a + ") with (" + dd + ":" + d + ")");
	            //remove(ad, a,d, aa,dd);
	            //remove(da, d,a, dd,aa);
	        } else {
	            int best = getBestKey(ab,ac,ad);
	            logger.info("best key = " + best);
	            if (best == -1) {
	                a = IntUtil.getRandomInt(list.size());
	                logger.info("Random " + aa + ":" + a);
	                
	                // nothing to delete
	            } else {
	                logger.info("Best " + aa + ":" + best);
	                a = best;            
	                int[] vals = getRandomValues(a, ab,ac,ad);
	                int tmpb = vals[0];
	                int tmpc = vals[1];
	                int tmpd = vals[2];
	                
	                if (b == -1 && tmpb != -1) {
	                	logger.info("New " + bb + " = " + tmpb);
	                	b = tmpb;
	    	            //remove(ab, a,b, aa,bb);
	    	            //remove(ba, b,a, bb,aa);
	                }//b
	                if (c == -1 && tmpc != -1) {
	                	logger.info("New " + cc + " = " + tmpc);
	                    c = tmpc;
	    	            //remove(ac, a,c, aa,cc);
	    	            //remove(ca, c,a, cc,aa);
	                }//c
	                if (d == -1 && tmpd != -1) {
	                	logger.info("New " + dd + " = " + tmpd);
	                    d = tmpd;
	    	            //remove(ad, a,d, aa,dd);
	    	            //remove(da, d,a, dd,aa);
	                }//d		                
	            }//else
	        }//else
        } // a == -1
    	
	    return new int[] { a, b, c, d };
    }	// doIt
    
    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------

    /**
     * 
     * @param key
     * @param map
     * @return
     */
    private static int getRandomValue(Integer key, Map<Integer, List<Integer>> map) {
    	int val = -1;
    	
    	if (key!=null && map!=null && map.get(key)!=null && 0<map.get(key).size()) {
    		return ListUtil.getRandom(map.get(key)).intValue();
    	}
    	
    	return val;
    }
    
    private static int[] getRandomValues(Integer key, Map<Integer, List<Integer>> map, Map<Integer, List<Integer>> map2, Map<Integer, List<Integer>> map3) {
    	int[] vals = new int[3];
    	vals[0] = getRandomValue(key, map);
    	vals[1] = getRandomValue(key, map2);
    	vals[2] = getRandomValue(key, map3);
    	
    	return vals;
    }
    
    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------

    /**
     * 
     * @param map
     * @param key
     * @param value
     */
    private static void remove(Map<Integer, List<Integer>> map, Integer key, Integer value, String a, String b) {
    	if (map==null || key==null || value==null) {
    		return;
    	}
    	
    	//
    	// remove key value pair
    	List<Integer> list = map.get(key);
    	if (list!=null && 0<list.size()) {
        	int i = list.indexOf(value);
        	if (0<=i) { 
        		printMap("Before: " + a+b, map);
        		list.remove(i); 
        		logger.info("Removing k/v (" + key + "/" + value + ") from '" + a + b + "' map.");
        	}
    	}
    	
    	//
    	//
    	if (list==null || list.size()==0) {
    		map.remove(key);
    		logger.info(a + b + " map has no more entries for key: " + key);
    	}
		printMap("After: " + a+b, map);
    	
    	//
    	//
    	if (map.size()==0) {
    		map=null;
    		logger.info(key + " was the last key in map '" + a + b + "'");
    	}
    } //remove 
    
    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------

    /**
     * 
     * @param map
     * @param map2
     * @param map3
     * @return
     */
    private static int getBestKey(Map<Integer, List<Integer>> map, Map<Integer, List<Integer>> map2, Map<Integer, List<Integer>> map3) {
    	int best = -1;
    	
    	if (map!=null && 0<map.size()) {
    		for (Integer i : map.keySet()) {
    			return i;
    		}
    	}
    	
    	if (map2!=null && 0<map2.size()) {
    		for (Integer i : map2.keySet()) {
    			return i;
    		}
    	}

    	if (map3!=null && 0<map3.size()) {
    		for (Integer i : map3.keySet()) {
    			return i;
    		}
    	}

    	return best;
    }

    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------

    /**
     * 
     * @param actualRows
     * @param indexes
     * @return
     */
    private static int[][] removeEmptyRows(int actualRows, int[][] indexes) {
    	int cols = indexes[0].length;
    	int[][] newIndexes = new int[actualRows][cols];
    	for (int a=0; a<actualRows; a++) {
    		newIndexes[a] = indexes[a];
    	}
    	return newIndexes;
    }
    
    // --------------------------------------------------------
    // ----- 
    // --------------------------------------------------------
    
    /**
     * 
     * @param mapname
     * @param map
     */
    private static void printMap(String mapname, Map<Integer, List<Integer>> map) {
    	if (map==null || map.size()==0) { return; }
    	StringBuffer sb = new StringBuffer();
    	sb.append(mapname);
    	for (Integer i : map.keySet()) {
    		List<Integer> list = map.get(i);
    		sb.append("\t" + i + "->" + ListUtil.getCommaDelimitedStr(list) + ",");
    	}
    	logger.info(sb.toString());
    }
    
} // AllPairsHelper


		





